/****************************************************************/
/*                                                              */
/*                      WICO2XPM                                */
/*                      --------                                */
/*                   Version Beta 0.91                          */
/*  Conversion d'icones Windows (tm) en fichiers XPM pour       */
/*                         X-Window                             */
/*                                                              */
/****************************************************************/

/*
	Par Serge Delbono (sdelbono@mygale.org)
	Copyright (C) 1995 Serge Delbono

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ATTENTION ! WARNING ! */
/* Dans cette version, n'utiliser que sur une machine "Big Endian" */
/* comme les 80x86 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#ifndef SEEK_SET
#define SEEK_SET 0
#endif
/* #include <X11/xpm.h> */

/* structure des fichiers ICO de Windows */

/* types Microsoft */

#define FALSE               0
#define TRUE                1
typedef unsigned short WORD ;
typedef unsigned char BYTE ;
typedef unsigned long DWORD ;
typedef long LONG ; /* c'est malin ! */

/* cECi EsT unE NoTAtIoN hONgRoIse (lIsIblE iSn't It !) */

typedef struct tagBITMAPINFOHEADER
{
    DWORD   biSize;
    long    biWidth;
    long    biHeight;
    WORD    biPlanes;
    WORD    biBitCount;
    DWORD   biCompression;
    DWORD   biSizeImage;
    long    biXPelsPerMeter;
    long    biYPelsPerMeter;
    DWORD   biClrUsed;
    DWORD   biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagRGBQUAD
{
    BYTE    rgbBlue;
    BYTE    rgbGreen;
    BYTE    rgbRed;
    BYTE    rgbReserved;
} RGBQUAD;

typedef struct IconDirectoryEntry {
	BYTE bWidth ;
	BYTE bHeight ;
	BYTE bColorCount ;
	BYTE bReserved ;
	WORD wPlanes;
	WORD wBitCount ;
	DWORD dwByteInRes ;
	DWORD dwImageOffset ;
} ICONDIRENTRY ;

typedef struct ICONDIR {
	WORD idReserved ;
	WORD idType ;
	WORD idCount ;
	ICONDIRENTRY idEntries[1] ;
} ICONHEADER ;



/* declaration des fonctions */
void main (int argc, char *argv[]) ;
void usage (char *myname) ;
void traiter (char *fichier) ;
void traiterimage (FILE *fi, FILE *fo) ;

ICONDIRENTRY entree ;


	/********************************/
	/*	MAIN ROUTINE		*/
        /********************************/

void main (int argc, char *argv[])
{
	int i ;
	if (argc <=1) {
		usage(argv[0]) ;
		exit (1) ;
	}
	for (i=1; i<argc; i++)
		traiter (argv[i]) ;
	exit (0) ;
}

void usage (char *myname)
{
	printf ("Conversion d'icnes Windows en XPM\n") ;
	printf ("par Serge Delbono\n\nUsage :") ;
	printf ("%s <liste de fichiers icones>\n",myname) ;
}
	
	/********************************/
	/*   TRAITEMENT d'UN FICHIER	*/
	/*******************************/

void traiter (char *fichier)
{
	FILE *fi,* fo ;
	DWORD magic ;
	short idCount ;
	int numimage ;
	long curpos ;
	static char outfile[100] ;
	printf ("%s : ",fichier) ;
	fi = fopen (fichier,"rb") ;
	if (!fi)
	{
		perror ("On ne peut ouvrir le fichier") ;
		return ;
	}
	/* on decode le ICONHEADER */
	magic = 0xffffffff ;
	fread (&magic, 4, 1, fi) ;
	if (magic != 0x00010000)
	{
		printf ("Bad magic ! (%lx)\n",magic) ;
		fclose (fi) ;
		return ;
	}
	fread (&idCount,2,1,fi) ;
	printf ("%d image[s]\n",idCount ) ;
	
	for (numimage=0; numimage<idCount; numimage++)
	{
		if (fread(&entree,sizeof(entree),1,fi) !=1) {
			perror ("EOF in entree") ;
			fclose (fi) ;
			return ;
		}
		strcpy (outfile,fichier) ;
		if (strcmp(fichier+strlen(fichier)-4,".ico") == 0)
			*(outfile+strlen(fichier)-4)='\0' ;
		if (numimage)
			sprintf (outfile,"%s%d.xpm",outfile,numimage) ;
		else
			strcat (outfile,".xpm") ;
		fo = fopen (outfile,"w") ;
		printf (" %s : ",outfile) ;
		if (!fo)
		{
			perror ("cration fichier") ;
			fclose (fi) ;
			return ;
		}
		curpos = ftell (fi) ;
		traiterimage (fi,fo) ;
		fclose (fo) ;
		fseek (fi,curpos,SEEK_SET) ;
		
	}
	
	fclose (fi) ;	
}

void traiterimage (FILE *fi, FILE *fo)
{
	int width, height, ncolors ;
	int i,j ;
	RGBQUAD rgbin ;
/* struture pour memoriser la palette */	
	typedef struct {
		DWORD val; /* valeur RGB de l'entree */
		short used;/* entree utilisee ou pas ? */
	} PAL ;
	PAL *pal ;
	PAL invisible ;
	short *image, *pimage ;
	int nboct ;
	width = entree.bWidth ;
	height = entree.bHeight ;
	ncolors = entree.bColorCount ;
	printf ("%dx%d (%d colors)\n",width, height, ncolors) ;
	if (
		(width > 32) ||
		(width < 16) ||
		(height > 32)||
		(height <16) ||
		((ncolors != 16) &&
		(ncolors != 2) /* &&
		(ncolors != 8) */)
	   ) {
	   fprintf(stderr, "Taille/nb de couleurs incompatible\n") ;
	   return ;
	}
	fprintf (fo,"/* XPM */\nstatic char* pixmap = {\n") ;
	
	/* on saute vers les RGBQUADs */
	fseek (fi, entree.dwImageOffset+sizeof(BITMAPINFOHEADER),SEEK_SET) ;

/* on cree la palette : pour eviter les couleurs non utilisees,
	on memorise la palette */
	pal = malloc ((ncolors+1)*sizeof(PAL)) ;
	if (!pal)
	{
		fprintf(stderr,"Plus de place pour la palette !\n") ;
		return ;
	}

	for (i=0;i < ncolors; i++)
	{
		fread(&rgbin,sizeof(rgbin),1,fi) ;
		(pal+i)->val = (DWORD)rgbin.rgbBlue +
			(((DWORD)rgbin.rgbGreen)<<8) +
			(((DWORD)rgbin.rgbRed)<<16) ;
		(pal+i)->used = FALSE ;
	}
/* et une entree pour la couleur "invisible" */	
	invisible.val = 0 ;
	invisible.used = FALSE ;
	image = (short *)malloc(width * height * sizeof (short)) ;
	if (!image)
	{
		fprintf (stderr,"Plus de place pour l'image !\n") ;
		free (pal) ;
		return ;
	}
	pimage = image ;
	
	switch (ncolors)
	{
		case 2 : /* monochrome icon */
		 nboct = (width * height) / 8;
		 for (i=0;i < nboct; i++)
		 {
			BYTE c = fgetc(fi) ;
			for(j=0;j<8;j++)
			{
				WORD val ;
				val = (c & 0x80) >> 7 ;
				(pal+val)->used = TRUE ;
				*pimage = val ;
				c = c << 1 ;
				pimage ++ ;
			}
			
		 }
		 break ;
		
		case 8: /* 4 couleurs : c'est la m**** */
		fprintf(stderr,"Dsol : 8 couleurs non support...\n") ;
		
		 break ;
		
		case 16: /* 16 couleurs, 2 pixels par octets */
		 nboct = (width * height) / 2 ;
		 for (i=0;i<nboct;i++)
		 {
			BYTE c = fgetc(fi) ;
			for (j=0;j<2;j++)
			{
				WORD val ;
				val = (c & 0xf0) >>4 ;
				*pimage = val ;
				(pal+val)->used = TRUE ;
				c = c << 4 ;
				pimage ++ ;
			}
		 }
		 break ;

	} /* fin switch ncolors */
	
	/* derriere, on a la table des AND. IE un champs de bits des zones
	"transparentes" */
	pimage = image ;
	nboct = (width * height)/8 ;
	for (i=0; i < nboct; i++)
	{
		BYTE c = fgetc(fi) ;
		for (j=0;j<8;j++)
		{
			if ((c & 0x80) != 0)
			{
			   *pimage = -1 ;
			   invisible.used = TRUE ;
			}
			pimage ++ ;
			c = c<<1;
		}		
	}
	
/* ecriture des donnees */	
/* ecriture de la palette */
	j = 0 ;
	for (i=0;i<ncolors;i++)
	{
		if ((pal+i)->used)
		  j ++ ;
	}
	if (invisible.used) j++ ;
	fprintf(fo,"\"%d %d %d 1\",\n",
		width,height, j) ;
	if (invisible.used)
	fprintf(fo,"\"  s None  c None\",\n") ;	
	for (i=0;i<ncolors;i++)
	if ((pal+i)->used)
	{
                fprintf(fo,"\"%c c #%06lx\",\n",'a'+i,(pal+i)->val) ;
	}
	free(pal) ;
/* ecriture de l'image */
	for (i=0;i<height; i++)
	{
		short c ;
		putc('\"',fo) ;
/* Telles Windows, les images BMP et ICO sont "tete en bas" */
/* ie : la premiere ligne de la bitmap est en bas de l'image */
		pimage = image + (height-i-1)*width ;
		for (j=0;j<width;j++)
		{
			c = *pimage ;
			pimage ++ ;
			if (c == -1)
				putc(' ',fo) ;
			else
				putc('a'+c,fo) ;
		}
		if (i==(height-1)) /* derniere ligne */
			fprintf(fo,"\"};\n") ;
		else
			fprintf(fo,"\",\n") ;
	}
	free(image) ;
}
